# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Raw README plugin for html renderer.

.. autoclass:: HtmlReadmeRenderer
   :members:

"""

from ...jinja2.readme import Jinja2ReadmeRenderer


class HtmlReadmeRenderer(Jinja2ReadmeRenderer):
    """Html renderer for readme files, using jinja2 template engine.

    It uses the row content of the ``html`` README.
    """

    # pylint: disable=too-few-public-methods, abstract-method

    plugin_type = "renderer.html.readme"
    keyword = "renderer.html.readme.html"
    extensions = ["html", "htm"]
