{% macro file(path) -%}
  <img class="thumbnail" src="{{ local.setup['href_prefix'] }}{{ path }}" height="10" width="10" alt="{{ path.as_posix()|basename }}" >
  <a href="{{ local.setup['href_prefix'] }}{{ path }}">
    {{ path.as_posix()|basename }}
  </a>
{%- endmacro %}
